// ==========================================================
//
// CBP_Hexapod_2.c   25-12-21
//
// Hexapod robot following my structure
//
// CBP 05-12-20
// Rev. 08-12-20
// 1) Calc of rand number corrected - typecast to double
// 2) Inits of min, max, offset taken out of loop
// 3) omega = 3*pi
// 4) Give name to octave file for each robot
//
// CBP 04-03-21 Cleanup, Added Octave data capture and auto-plot
// CBP 25-12-21  Replace min/max/offset caculation of leg drive
// with a generic sigmoid.
// ==========================================================
 
#include <stdio.h>
#include <webots/motor.h>
#include <webots/robot.h>
#include <webots/gps.h>
#include <stdlib.h>

// ODE Solvers
#include "rkf45.h"
#include "CBP_rhs_hexapod.h"
#include "CBP_Helper_H.h"

#define TIME_STEP 16
#define NUM_MOTORS 12

#define NEQN 23
 
#define UP -0.015
#define DOWN 0.015

#define DEBUG false

// defines for the ODEs for the feet
#define LF  0
#define LM  4
#define LR  7
#define RF	10
#define RM	13
#define RR	16

// defines for the ODEs for the legs
#define LF_l 17
#define LM_l 18
#define LR_l 19
#define RF_l 20
#define RM_l 21
#define RR_l 22

double threshold(double phase, double thresh){
  if(phase > thresh)
    return phase;
  else 
    return 0.0;
}

double setFoot(double phase, double thresh){
  if(phase > thresh)
    return UP;
  else 
    return DOWN;
}

double sigmoid(double in) {
  return 1.0/(1 + exp(-in));
}

 
int main(int argc, char **argv) {

  // rkf stuff
  double abserr;
  int flag;
  double relerr;
  double t;
  double t_out;
  double y[NEQN];
  double dydt[NEQN];
  
  double pi = M_PI;
  
  // params
  double alpha = 5;
  double mu = 1;
  //double omega = 3.0*pi;
  double omega = pi;
  double phase;
  double thresh;
  double tau,k;
  double scaling;
  
  // state variables
  double phaseLF, phaseLM,phaseLR,phaseRF,phaseRM,phaseRR;
  double phaseLF_l, phaseLM_l,phaseLR_l,phaseRF_l,phaseRM_l,phaseRR_l;
 
  // GPS stuff for speed calculation
  double x=0, xPrev, yy=0, yyPrev, speed;
  double avSpeed;
  int count=0;
  const char *name;  
  char fname[10];
  
  double phaseArray[12];
  double gaitAngle;
  
  FILE* logFile;
  FILE* exptFile;
  
  const char *MOTOR_NAMES[NUM_MOTORS] = {"hip_motor_r0",  "hip_motor_r1",  "hip_motor_r2",  "hip_motor_l0",
                                         "hip_motor_l1",  "hip_motor_l2",  "knee_motor_r0", "knee_motor_r1",
                                         "knee_motor_r2", "knee_motor_l0", "knee_motor_l1", "knee_motor_l2"};
  WbDeviceTag motors[NUM_MOTORS];
  WbDeviceTag gps;
  
  int i;

  wb_robot_init();
  
  gaitAngle = atof(argv[1]);
   
  // Get robot name
  name = wb_robot_get_name();
  sprintf(fname,"Robot_%s.m",name);
  logFile = fopen(fname,"w");
  
  writeOctaveHead(logFile);
  
  sprintf(fname,"%s.txt",name);
  printf("%s\n",fname);
  exptFile = fopen(fname,"a");
  
  for (i = 0; i < NUM_MOTORS; i++) {
    motors[i] = wb_robot_get_device(MOTOR_NAMES[i]);
    if (!motors[i])
      printf("could not find motor: %s\n", MOTOR_NAMES[i]);
  }
  
  // GPS for speed calculations
  gps = wb_robot_get_device("gps");
  wb_gps_enable(gps, TIME_STEP); 
  avSpeed = 0;  

 // Calculate the epsilon
  abserr = 100*sqrt ( r8_epsilon ( ) );
  relerr = 100*sqrt ( r8_epsilon ( ) );
  printf("Root epsilon: %e\n",abserr);

  // Set some initial conditions
  for(i=0; i < NEQN; i++)
    y[i] = 0.0;
  y[0] = 0.7;
  y[1] = 0.7;

  
  // Choose the gait ----------------------------------------
  phase = gaitAngle*pi/180;
  tau = tan(phase/3.0)/omega;
  k = sqrt(tau*tau*omega*omega + 1);
  thresh = sin(pi/2 - 0.5*phase);

  fprintf(exptFile,"----------------------------------------\n"); 
  fprintf(exptFile,"omega  = %f\n",omega);  
  fprintf(exptFile,"phase  = %f (%2f)\n",phase,phase*180/M_PI);
  fprintf(exptFile,"tau    = %f\n",tau);
  fprintf(exptFile,"k      = %f\n",k);
  fprintf(exptFile,"thresh = %f\n",thresh);     
  fprintf(exptFile,"\n"); 
  
  setParams(alpha, mu, omega,tau,k,thresh);
	
  scaling = 0.5;
  
  printf("thresh = %f\n",thresh);
   
  // Initialize the ODE Solver                 
  t = 0.0;
  t_out = TIME_STEP/1000.0; 
  flag = 1;
    
  while (wb_robot_step(TIME_STEP) != -1) {
  
                      
    flag = r8_rkf45 ( rhs_hexapod, NEQN, y, dydt, &t, t_out, &relerr, abserr, flag );
       
    t_out = t + TIME_STEP/1000.0;
    
    if(flag == 4) {
      printf("WARNING: flag 4 at t = %f\n",t);
      flag = 2;
    }
    if(flag == 7) {
      printf("WARNING: flag 7 at t = %f\n",t);
      flag = 2; 
    }   
    if(flag == 6) {
      abserr *=1.1;
      relerr *=1.1;
      flag = 2;
      printf("WARNING: flag 6: abserr = %f\n",abserr);
    }
    
    phaseLF = y[LF];
    phaseRF = y[RF];
    phaseLM = y[LM];
    phaseRM = y[RM];	
    phaseLR = y[LR];
    phaseRR = y[RR];      
 
    phaseLF_l = 4*sigmoid(y[LF_l]) - 3;
    phaseRF_l = 4*sigmoid(y[RF_l]) - 3;
    phaseLM_l = 4*sigmoid(y[LM_l]) - 3;
    phaseRM_l = 4*sigmoid(y[RM_l]) - 3;
    phaseLR_l = 4*sigmoid(y[LR_l]) - 3;
    phaseRR_l = 4*sigmoid(y[RR_l]) - 3;

    phaseArray[0] = y[LF];   phaseArray[1]  = y[RF]; phaseArray[2] = y[LM];
    phaseArray[3] = y[RM];   phaseArray[4]  = y[LR]; phaseArray[5] = y[RR];
    phaseArray[6] = phaseLF_l; phaseArray[7]  = phaseRF_l; phaseArray[8] = phaseLM_l;
    phaseArray[9] = phaseRM_l; phaseArray[10] = phaseLR_l; phaseArray[11] = phaseRR_l;
            
    writeDataAllToOctave(logFile,t,phaseArray);

  if(t > 10) {

    // Move the legs
    wb_motor_set_position(motors[2],-phaseLF_l*scaling);
    wb_motor_set_position(motors[5], phaseRF_l*scaling);
    wb_motor_set_position(motors[1],-phaseLM_l*scaling);
    wb_motor_set_position(motors[4], phaseRM_l*scaling);
    wb_motor_set_position(motors[0],-phaseLR_l*scaling);
    wb_motor_set_position(motors[3], phaseRR_l*scaling);
       
    // Move the feet
    wb_motor_set_position(motors[8], setFoot(phaseLF,thresh)); 
    wb_motor_set_position(motors[11],setFoot(phaseRF,thresh));
    wb_motor_set_position(motors[7], setFoot(phaseLM,thresh));
    wb_motor_set_position(motors[10],setFoot(phaseRM,thresh));
    wb_motor_set_position(motors[6], setFoot(phaseLR,thresh));
    wb_motor_set_position(motors[9], setFoot(phaseRR,thresh));
    
  }
     
    // GPS for speed calc
    xPrev = x;
    yyPrev = yy;
    const double *gpsVals = wb_gps_get_values(gps);  
    x = gpsVals[0];
    yy = gpsVals[2];    
    speed = sqrt(pow(x - xPrev,2) + pow(yy - yyPrev,2))/(TIME_STEP/1000.0);

    
    if(t > 10) {
      avSpeed += speed;
      count++;
      if(count > 100) {
        avSpeed = avSpeed/(double)count;
        printf("%s Average speed = %f\n",name,avSpeed);
        fprintf(exptFile," t = %3f, avSpeed = %f\n",t,avSpeed);
        count = 0;
        avSpeed = 0;
      }
    }  
	 
 } 
 
  writeOctaveTail(logFile,thresh);
  fflush(logFile);
  fclose(logFile);
  fflush(exptFile);
  fclose(exptFile);

  wb_robot_cleanup();

  return 0;
}
